"""
This module contains the assertions and error definitions.


"""
from schul_cloud_resources_server_tests.tests.assertions import assertIsError

ERROR_CLIENT_REQUEST = 409 # https://httpstatuses.com/409
Q = "Q" # The query string



"""
This file contains the test configuration like fixtures.
"""
from pytest import fixture
from schul_cloud_resources_server_tests.tests.fixtures import ParallelBottleServer
from bottle import Bottle, request
import sys
import os
import requests
from urllib.parse import urlencode
import copy


HERE = os.path.dirname(__file__)
MODULE_ROOT = os.path.join(HERE, "..", "..")
try:
    import schul_cloud_search_tests
except ImportError:
    sys.path.append(MODULE_ROOT)
from schul_cloud_search_tests.proxy import get_app


class Requester(object):
    """Shortcut to request a hosted resource."""
    
    def __init__(self, url):
        """Create a requester."""
        self._url = url
    
    def request(self):
        """Request the resource."""
        return requests.get(self._url)


def ending_with_slash(url):
    """Return the url with a slash in the end."""
    if not url.endswith("/"):
        url = url + "/"
    return url


def params_to_key(params):
    """Return a tuple key for the parameters."""
    key = list(params.items())
    key.sort()
    return tuple(key)

class SearchEngine(object):
    """The search engine adapter.
    
    This includes:
    - the running search tests server
    - a bottle server returning the registeres responses
    """
    
    _default_response = {
      "jsonapi": {
        "version": "1.0",
        "meta" : {
          "name": "Example Server",
          "source": "https://github.com/schul-cloud/resources-api-v1",
          "description": "This is just an eampel server for the search API."
        }
      },
      "links": {
        "self": {
          "href": None,
          "meta": {
            "count": 0,
            "offset": 0,
            "limit": 0
          }
        },
        "first": None,
        "last": None,
        "prev": None,
        "next": None
      },
      "data": []
    }

    def __init__(self):
        """Create a search engine object."""
        self._queries = {}
        self._app = Bottle()
        self._app.get("/", callback=self._serve_request)
        self._reset_servers()
    
    def start(self):
        """Start the search engine tests and the search engine mock."""
        assert not self.is_started()
        self._server = ParallelBottleServer(self._app)
        self._search_app = get_app(target_url=self.search_engine_url)
        self._search_server = ParallelBottleServer(self._search_app)
        
    @property
    def search_engine_url(self):
        """Return the URL of the mocked search engine."""
        assert self.is_started()
        return ending_with_slash(self._server.url)

    @property
    def proxy_url(self):
        """Return the url of the search engine tests proxy."""
        assert self.is_started()
        return ending_with_slash(self._search_server.url)
        
    def stop(self):
        """Stop the search engine tests and the mock server."""
        assert self.is_started()
        self._server.shutdown()
        self._search_server.shutdown()
        self._reset_servers()

    def _reset_servers(self):
        """Create the starting condition"""
        self._server = None
        self._search_server = None
        self._search_app = None

    def is_started(self):
        """Return whether the server is started."""
        return self._server is not None
        
    def host(self, response, params={}):
        """Host the response given by the query."""
        key = params_to_key(params)
        assert key not in self._queries
        self._queries[key] = response
        return self._new_requester(params)
    
    def clear(self):
        """Remove all hosted responses."""
        self._queries = {}
        self._last_response = None

    def _serve_request(self):
        """Serve a request to the search engine bottle server."""
        default = self.get_default_response(request.query_string)
        response = self._queries.get(params_to_key(request.query), default)
        self._last_response = response
        return response

    def request(self, params={}):
        """Request a search with parameters."""
        return self._new_requester(params).request()
        
    def _new_requester(self, params):
        """Return a new Requester object from the parameters."""
        request_url = self.proxy_url + "?" + urlencode(params)
        return Requester(request_url)
        
    def get_default_response(self, query_string):
        """Return the default response to a query."""
        result = copy.deepcopy(self._default_response)
        url = self.search_engine_url + "?" + query_string
        result["links"]["self"]["href"] = url
        return result
        
    @property
    def last_response(self):
        """Return the last reponse of the search engine to a request.
        
        If there was no request, None is returned.
        """
        return self._last_response


@fixture(scope="session")
def search_engine_session():
    """Return the server to store resources."""
    server = SearchEngine()
    server.start()
    yield server
    server.stop()


@fixture
def search_engine(search_engine_session):
    """Return a fresh server object with no resources."""
    search_engine_session.clear()
    yield search_engine_session
    search_engine_session.clear()


@fixture
def code_url(search_engine):
    """Return the url of the code enpoint."""
    return search_engine.proxy_url + "code"


"""
Each test function here covers a part of the specification.
They are run when the server returns a request.
If they fail, the client gets feedback.

About the test functions:
Please provide
- A short description of what is tested
- Links to the specification, where this is specified.

"""
from pytest import mark
from schul_cloud_resources_api_v1.schema import get_schemas


@mark.skip(reason="TODO")
def test_response_schema(result):
    """If the request was successful, it must match the search-response schema.
    
    Schema:
    - https://github.com/schul-cloud/resources-api-v1/tree/master/schemas/search-response
    
    The response is successful of a status code 200 is returned.
    """


@mark.skip(reason="TODO")
def test_400_and_500_status_codes_have_the_jsonapi_design():
    """If the status code is 4XX, it contains a list of errors.

    Schema:
    - https://github.com/schul-cloud/resources-api-v1/tree/master/schemas/error
    """


@mark.skip(reason="TODO")
def test_the_content_type_is_from_the_jsonapi():
    """Make sure the returned content type is specified and expected.
    
    See here:
    - http://jsonapi.org/format/#content-negotiation-clients
    """



"""
Each test function here covers a part of the specification.
They are run when the server returns a request.
If they fail, the client gets feedback.

About the test functions:
Please provide
- A short description of what is tested
- Links to the specification, where this is specified.

These test functions all relate to the "links" attribute of the search response.
- https://github.com/schul-cloud/resources-api-v1/tree/master/schemas/search-response#search-reponse-1

"""
from pytest import mark


@mark.skip(reason="TODO")
def test_self_link():
    """The self link is has the same offset and may have a reduced limit.
    """


@mark.skip(reason="TODO")
def test_limit_may_be_reduced():
    """If the limit is higher than the maximum limit the server has internally, the server sets the limit to the maximum available limit.
    """


@mark.skip(reason="TODO")
def test_count():
    """count is the actual number of objects retrieved.
    
    The count must be equal to the number of objects in the list.
    """


@mark.skip(reason="TODO")
def test_offset():
    """offset is the start index in the list of objects.
    
    The offset must be the same as requested.
    """


@mark.skip(reason="TODO")
def test_limit():
    """limit is the requested number of objects.

    
    """


@mark.skip(reason="TODO")
def test_the_end_is_reached():
    """If the end of the resource list is reached, count may be less than limit.
    
    This implies that there is no next link and the last link is self.
    """


@mark.skip(reason="TODO")
def test_no_resources_given():
    """If there are no resources, first, last, prev and next MUST be null.
    """


@mark.skip(reason="TODO")
def test_last_page_is_reached():
    """If the last page is reached, next MUST be null.
    
    This is tested by requesting the next page and getting no resources back.
    """


@mark.skip(reason="TODO")
def test_last_implies_next():
    """If last is given, next must be given.
    """


@mark.skip(reason="TODO")
def test_first_implies_prev():
    """If first is given, prev must be given.
    """


@mark.skip(reason="TODO")
def test_do_not_skip_objects():
    """The prev and next links MUST not skip objects.
    
    This can only be inferred by the limit and offset.
    """


@mark.skip(reason="TODO")
def test_retrieve_object_with_self_link():
    """If the object has a self link, it can be retrieved.
    
    This must be tested in parallel to reduce the waiting time.
    """


@mark.skip(reason="TODO")
def test_links_are_absolute():
    """All links are absolute links using the host header field.
    """


@mark.skip(reason="TODO")
def test_request_too_far():
    """Requesting beyond the last link yields no objects.
    """



from schul_cloud_resources_api_v1.schema import get_schemas
from pytest import mark
from schul_cloud_search_tests.tests.assertions import (
    assertIsError, ERROR_CLIENT_REQUEST, Q)
from pprint import pprint


@mark.parametrize("response", get_schemas()["search-response"].get_valid_examples())
def test_valid_query_is_returned(search_engine, response):
    result = search_engine.host(response, {Q:"test"}).request()
    data = result.json()
    pprint(data)
    assert data == response


@mark.skip(reason="TODO")
def test_invalid_return_values_are_handled():
    """Test that a invalid responses still create a useful output.
    
    - invalid json
    - valid json but too small
    - bytes
    """


@mark.skip(reason="TODO")
def test_invalid_return_header_type_is_handled():
    """The header must be correct.
    
    Wrong headers:
    - with parameters
    - application/json
    """


